# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import os

from faststream.rabbit import RabbitBroker
from bovine_process.types import ProcessingItem

from .exchanges import processing
from .types import ProcessingMessage

amqp_uri = os.environ.get("BOVINE_AMQP", "amqp://localhost")

enqueue_broker = RabbitBroker(amqp_uri)
"""Broker to be used for enqueuing items. amqp uri is determined from BOVINE_AMQP
environment variable. Otherwise `amqp://localhost`"""


async def enqueue_to_inbox(item: ProcessingItem, actor):
    """Enqueues a message for inbox processing"""
    await enqueue_broker.publish(
        ProcessingMessage(
            bovine_name=actor.bovine_name, data=item.data, submitter=item.submitter
        ),
        exchange=processing,
        routing_key="inbox",
    )


async def enqueue_to_outbox(item: ProcessingItem, actor):
    """Enqueues a message for outbox processing"""
    await enqueue_broker.publish(
        ProcessingMessage(
            bovine_name=actor.bovine_name, data=item.data, submitter=item.submitter
        ),
        exchange=processing,
        routing_key="outbox",
    )



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from bovine_process.utils import ProcessorList, ByActivityType

from bovine_process.common import interaction_processor, sanitize

from .following import accept_follow, undo_follow
from .outgoing_delete import outgoing_delete
from .outgoing_update import outgoing_update
from .store_outgoing import add_outgoing_to_outbox, store_outgoing
from .update_id import update_id

default_outbox_process = ProcessorList(
    sanitize,
    update_id,
    store_outgoing,
    add_outgoing_to_outbox,
    ByActivityType(
        Update=outgoing_update,
        Delete=outgoing_delete,
    ),
)
"""Defines the synchronous part of sending an outgoing object"""


social_processor = ByActivityType(Accept=accept_follow, Undo=undo_follow)
"""Processes outgoing social events, i.e. updates the social graph"""


default_async_outbox_process = ProcessorList(social_processor, interaction_processor)
"""Defines the asynchronous part of handling an outgoing object"""



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import logging
from abc import ABC, abstractmethod

from bovine.activitystreams.utils import id_for_object
from bovine.parse import Activity
from bovine_process.types import ProcessingItem

logger = logging.getLogger(__name__)


class InteractionActor(ABC):
    """Abstract class to illustrate what is needed to process interactions"""

    @abstractmethod
    async def retrieve(self, object_id: str, only_own: bool = False) -> dict | None:
        """Retrieves an object

        :param object_id: id of the object
        :param only_own: if true requires object to be by the actor"""
        pass

    @abstractmethod
    async def remove_references(self, object_id: str) -> None:
        """Removes references to the object. Used in case an object is
        deleted and it is unknown which object it replies to"""
        pass

    @abstractmethod
    async def add_to_interaction(
        self, interaction_type: str, object_id: str, interaction: str
    ) -> None:
        """Adds to interaction

        Example object for `interaction_type = "likes"`

        ```json
        {
            "type": "Like",
            "id": interaction,
            "object": object_id
        }
        ```

        :param interaction_type: One of `likes`, `shares, `replies`
        :param object_id: id of the object being interacted with
        :param interaction: The id of the interacting object
        """
        pass

    @abstractmethod
    async def remove_from_interaction(
        self, interaction_type: str, object_id: str, interaction: str
    ) -> None:
        """Removes from interaction

        :param interaction_type: One of `likes`, `shares, `replies`
        :param object_id: id of the object being interacted with
        :param interaction: The id of the interacting object"""

        pass


async def own_object(item: ProcessingItem, actor: InteractionActor) -> dict | None:
    object_id = id_for_object(item.data.get("object"))
    if object_id is None:
        return None
    return await actor.retrieve(object_id, only_own=True)


async def like_handler(item: ProcessingItem, actor: InteractionActor) -> ProcessingItem:
    """Adds object to the likes collection of the liked object, if

    * object being liked is owner by the receiving actor"""

    obj = await own_object(item, actor)
    if obj:
        obj_id = id_for_object(obj)
        if obj_id is None:
            return item
        logger.info("Like Handler %s", obj_id)
        await actor.add_to_interaction("likes", obj_id, item.data.get("id"))

    return item


async def announce_handler(
    item: ProcessingItem, actor: InteractionActor
) -> ProcessingItem:
    """Adds object to the shares collection of the announced object, if

    * object being announced is owner by the receiving actor"""

    obj = await own_object(item, actor)
    if obj:
        obj_id = id_for_object(obj)
        if obj_id is None:
            return item
        logger.info("Announce Handler %s", obj_id)
        await actor.add_to_interaction("shares", obj_id, item.data.get("id"))

    return item


async def reply_handler(
    item: ProcessingItem, actor: InteractionActor
) -> ProcessingItem:
    """Adds object to the replies collection. Object being replied to
    is determined from `inReplyTo`. Reply is added if the object
    belongs to the receiving actor."""
    create = Activity(item.data, domain=item.submitter_domain)
    remote = await create.object_for_create(actor.retrieve)

    if not remote:
        return item

    if not remote.in_reply_to:
        return item

    obj = await actor.retrieve(remote.in_reply_to, only_own=True)

    if obj:
        obj_id = id_for_object(obj)
        if obj_id is None:
            return item
        logger.info("Reply Handler %s", obj_id)
        await actor.add_to_interaction("replies", obj_id, remote.identifier)

    return item


async def delete_reply_handler(
    item: ProcessingItem, actor: InteractionActor
) -> ProcessingItem:
    """If a reply is deleted, removes it from the replies collection"""

    remote = item.data.get("object")
    if not remote:
        return item

    if isinstance(remote, dict) and remote.get("type") == "Person":
        return item

    await actor.remove_references(remote)

    return item


async def undo_handler(item: ProcessingItem, actor: InteractionActor) -> ProcessingItem:
    """For an Undo of a Like, Dislike, Announce , they are removed from
    the appropriate collection."""

    object_to_undo = id_for_object(item.data.get("object"))
    if object_to_undo is None:
        return item

    obj = await actor.retrieve(object_to_undo)
    if obj is None:
        return item

    remote_actor = id_for_object(item.data.get("actor"))
    if obj.get("actor") != remote_actor:
        logger.error("Mismatching actor in undo from %s", remote_actor)
        return item

    obj_type = obj.get("type")
    if obj_type in ["Like", "Dislike", "http://litepub.social/ns#EmojiReact"]:
        logger.info("Undo Handler for Like of %s", obj.get("id"))
        await actor.remove_from_interaction("likes", obj.get("object"), obj.get("id"))
    elif obj_type == "Announce":
        logger.info("Undo Handler for Like of %s", obj.get("id"))
        await actor.remove_from_interaction("shares", obj.get("object"), obj.get("id"))

    return item


interaction_handlers = {
    **dict(
        Announce=announce_handler,
        Create=reply_handler,
        Delete=delete_reply_handler,
        Dislike=like_handler,
        Like=like_handler,
        Undo=undo_handler,
    ),
    "http://litepub.social/ns#EmojiReact": like_handler,
}
"""The handlers being called for interactions"""



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import aiohttp
import os
import json

from faststream import Context
from faststream.annotations import Logger

from faststream.rabbit import RabbitBroker, RabbitQueue


from bovine_process.types import ProcessingItem
from bovine_process import process_inbox_item
from bovine_process.outgoing import default_async_outbox_process
from bovine_process.send_item import (
    determine_recipients,
    send_to_inbox,
    get_inbox_for_recipient,
)
from bovine.types import ServerSentEvent, Visibility
from bovine.jsonld import with_external_context

from bovine_propan.actor import bovine_actor
from bovine_propan.types import ProcessingMessage, SendMessage, FetchObjectMessage
from bovine_propan.exchanges import processing, processed

broker = RabbitBroker(os.environ.get("BOVINE_AMQP", "amqp://localhost"))

inbox_queue = RabbitQueue("inbox", auto_delete=False, routing_key="inbox")
fetch_object_queue = RabbitQueue(
    "fetch_object", auto_delete=False, routing_key="fetch_object"
)
inbox_queue_event = RabbitQueue("inbox_event", auto_delete=False, routing_key="inbox")
outbox_queue = RabbitQueue("outbox", auto_delete=False, routing_key="outbox")
to_send_queue = RabbitQueue("to_send", auto_delete=False, routing_key="to_send")


@broker.subscriber(inbox_queue, processing)
async def inbox_handler(
    message: ProcessingMessage,
    logger: Logger,
    session: aiohttp.ClientSession = Context(),
):
    """Handles events coming to the inbox"""
    async with bovine_actor(message.bovine_name, session) as actor:
        item = ProcessingItem(message.submitter, message.data)
        await process_inbox_item(item, actor)


@broker.subscriber(inbox_queue_event, processing)
async def inbox_to_event(
    message: ProcessingMessage,
    logger: Logger,
    session: aiohttp.ClientSession = Context(),
):
    """Creates a server send event for inbox events"""
    async with bovine_actor(message.bovine_name, session) as actor:
        data_s = json.dumps(message.data)
        event = ServerSentEvent(data=data_s, event="inbox")

        await broker.publish(
            event.encode(),
            routing_key=actor.actor_object.event_source,
            exchange=processed,
        )


@broker.subscriber(outbox_queue, processing)
async def outbox_handler(
    message: ProcessingMessage,
    logger: Logger,
    session: aiohttp.ClientSession = Context(),
):
    """Handles elements to the outbox"""
    async with bovine_actor(message.bovine_name, session) as actor:
        item = ProcessingItem(message.submitter, message.data)
        await default_async_outbox_process(item, actor)

        recipients = await determine_recipients(item, actor)
        to_send = with_external_context(item.data)

        for recipient in recipients:
            await broker.publish(
                SendMessage(
                    recipient=recipient,
                    data=to_send,
                    bovine_name=message.bovine_name,
                ),
                routing_key="to_send",
                exchange=processing,
            )

        data_s = json.dumps(item.data)
        event = ServerSentEvent(data=data_s, event="inbox")

        if "database_id" in item.meta:
            event.id = item.meta["database_id"]

        actor_info = actor.actor_object.build(visibility=Visibility.OWNER)
        event_source = actor_info["endpoints"]["eventSource"]

        await broker.publish(
            event.encode(), routing_key=event_source, exchange=processed
        )


@broker.subscriber(to_send_queue, processing)
async def to_send_handler(
    message: SendMessage, logger: Logger, session: aiohttp.ClientSession = Context()
):
    """Retrieves inbox for recipient and sends data"""
    async with bovine_actor(message.bovine_name, session) as actor:
        inbox = await get_inbox_for_recipient(actor, message.recipient)
        if inbox:
            await send_to_inbox(actor, inbox, message.data)


@broker.subscriber(fetch_object_queue, processed)
async def fetch_object(
    message: FetchObjectMessage,
    logger: Logger,
    session: aiohttp.ClientSession = Context(),
):
    """Meant for RPC calls. Fetches object and returns it"""
    async with bovine_actor(message.bovine_name, session) as actor:
        return await actor.retrieve(message.object_id)



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import logging

from typing import Callable, Awaitable

from bovine.jsonld import with_bovine_context

from bovine_process.utils import ByActivityType
from bovine_process.types import ProcessingItem

from .interactions import interaction_handlers, InteractionActor

logger = logging.getLogger(__name__)


interaction_processor: Callable[
    [ProcessingItem, InteractionActor], Awaitable[ProcessingItem]
] = ByActivityType(**interaction_handlers)
"""Defines the processor that handles interactions with an object"""


async def sanitize(item: ProcessingItem, actor):
    """Applies the default bovine context"""
    item.data = with_bovine_context(item.data)

    if item.submitter != item.data.get("actor"):
        logger.error("Got wrong submitter for an activity %s", item.submitter)
        logger.error(item.data)
        # return

    return item



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from bovine_process.utils import ProcessorList, ByActivityType
from bovine_process.common import interaction_processor, sanitize

from .following import accept_follow, undo_follow
from .handle_update import handle_update
from .incoming_delete import incoming_delete
from .store_incoming import add_incoming_to_inbox, store_incoming


crud_handlers = dict(Update=handle_update, Delete=incoming_delete)
"""The handlers being called for CRUD operations"""

social_handlers = dict(Accept=accept_follow, Undo=undo_follow)
"""The handlers being called for social interactions, i.e. updating the social graph"""


default_inbox_process = ProcessorList(
    sanitize,
    store_incoming,
    ByActivityType(**crud_handlers),
    ByActivityType(**social_handlers),
    interaction_processor,
    add_incoming_to_inbox,
)
"""Represents the default process undertaken by an inbox item"""



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import logging

logger = logging.getLogger(__name__)


async def do_nothing(item, *args):
    return item


class ByActivityType:
    """Simplifies handling activity by type, usage:

    ```python
    by_type = ByActivityType(One=do_one, Two=do_two})

    by_type({"type": one}, arg1, arg2) # calls do_one({"type": one}, arg1, arg2)
    ```
    """

    def __init__(self, **kwargs):
        self.actions = kwargs

    async def __call__(self, item, *args):
        try:
            item_type = item.data.get("type")
            if item_type is None:
                return item
            if isinstance(item_type, str):
                if item_type in self.actions:
                    return await self.actions[item_type](item, *args)
                else:
                    return item
            elif isinstance(item_type, list):
                for single_type in item_type:
                    item = await self.actions[single_type](item, *args)
                return item
            else:
                return item

        except Exception as ex:
            logger.error("Something went wrong with %s during procession", repr(ex))
            logger.exception(ex)
            if item:
                logger.error(item.data)


class ProcessorList:
    """Creates a list of processors, usage:

    ```python
    processors = ProcessorList(my_processor,
        ByActivityType(Follow=my_follow_processor))
    await processors(item, args)
    ```"""

    def __init__(self, *processors):
        self.processors = list(processors)

    async def __call__(self, item, *arguments):
        working = item

        try:
            for processor in self.processors:
                working = await processor(working, *arguments)
                if not working:
                    return

            return working
        except Exception as ex:
            logger.exception(ex)
            raise ex



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from faststream.rabbit import RabbitExchange, ExchangeType

processing = RabbitExchange("processing", auto_delete=False, type=ExchangeType.TOPIC)
"""Exchange used for messages being processed.

Relevant routing_key are:

* `inbox`
* `outbox`
"""

processed = RabbitExchange("processed", auto_delete=False, type=ExchangeType.TOPIC)
"""Exchange used for messages having been processed, currently used to feed the EventSources"""



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import os
import aiohttp
import logging

from faststream import FastStream, Context
from faststream.annotations import ContextRepo

from bovine_store.config import tortoise_config
from tortoise import Tortoise

from bovine_propan.broker import broker

logging.basicConfig(level=logging.INFO)


app = FastStream(broker, title="BovineProcessor")
"""faststream app can be run using 

```bash
faststream run bovine_propan/__init__:app
```
"""


@app.on_startup
async def startup(context: ContextRepo):
    db_url = os.environ.get("BOVINE_DB_URL", "sqlite://bovine.sqlite3")
    await Tortoise.init(config=tortoise_config(db_url))
    await Tortoise.generate_schemas()

    session = aiohttp.ClientSession()
    context.set_global("session", session)


@app.on_shutdown
async def shutdown(session: aiohttp.ClientSession = Context()):
    await Tortoise.close_connections()
    await session.close()



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import pytest
from faststream.rabbit import TestRabbitBroker
from unittest.mock import MagicMock

from bovine_process.types import ProcessingItem

from .enqueuers import enqueue_to_inbox, enqueue_broker


@pytest.fixture
async def test_broker():
    async with TestRabbitBroker(enqueue_broker) as broker:
        yield broker


async def test_enqueue_inbox(test_broker):
    item = ProcessingItem("submitter", {})
    await enqueue_to_inbox(item, MagicMock(bovine_name="bovine_name"))



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import logging
import aiohttp

from bovine_store import BovineStore
from contextlib import asynccontextmanager

logger = logging.getLogger(__name__)

cache = {}


@asynccontextmanager
async def bovine_actor(bovine_name: str, session: aiohttp.ClientSession):
    """Yields the BovineStoreActor corresponding to the `bovine_name`.
    Usage:

    ```python
    async with bovine_actor(bovine_name, session) as actor:
        ...
    ```
    """
    if bovine_name not in cache:
        store = BovineStore(session=session)
        actor = await store.actor_for_name(bovine_name)
        cache[bovine_name] = actor

    try:
        yield cache[bovine_name]
    except Exception as e:
        logger.exception(e)



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from pydantic import BaseModel, Field


class ProcessingMessage(BaseModel):
    """Message being processed

    see source for field details"""

    bovine_name: str = Field(
        examples=["vanilla_1234"],
        description="Unique identifier of the actor in the database",
    )

    submitter: str = Field(description="Actor that submitted the item")

    data: dict | list = Field(
        description="ActivityPub object being processed", examples=[{"type": "Like"}]
    )


class SendMessage(BaseModel):
    """Message being send

    see source for field details"""

    bovine_name: str = Field(
        examples=["vanilla_1234"],
        description="Unique identifier of the actor in the database",
    )

    recipient: str = Field(description="Actor the message is being send to")

    data: dict | list = Field(
        description="ActivityPub object being send", examples=[{"type": "Like"}]
    )


class FetchObjectMessage(BaseModel):
    """Request to fetch object"""

    bovine_name: str = Field(
        examples=["vanilla_1234"],
        description="Unique identifier of the actor in the database",
    )

    object_id: str = Field(description="The object to fetch")



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import aiohttp
import pytest

from faststream.rabbit import TestRabbitBroker
from faststream.annotations import ContextRepo
from unittest.mock import AsyncMock, MagicMock

from .actor import cache

from .broker import broker, inbox_queue, fetch_object_queue
from bovine_propan.exchanges import processing, processed


@pytest.fixture
async def test_broker():
    async with TestRabbitBroker(broker) as br:
        context = ContextRepo()
        context.set_global("session", AsyncMock(aiohttp.ClientSession))
        context.set_global("logger", AsyncMock())

        actor = AsyncMock()
        actor.actor_id = "http://local/actor"
        cache["bovine_name"] = actor
        actor.actor_object.build = MagicMock(
            return_value={"endpoints": {"eventSource": "xxxx"}}
        )
        actor.retrieve_own_object.return_value = MagicMock()
        actor.retrieve.return_value = {"some": "object"}

        yield br


async def test_inbox_handler(test_broker):
    remote_actor = "https://remote.example/actor"

    await test_broker.publish(
        {
            "bovine_name": "bovine_name",
            "submitter": remote_actor,
            "data": {
                "@context": "about:bovine",
                "actor": remote_actor,
                "type": "Like",
            },
        },
        exchange=processing,
        queue=inbox_queue,
    )


async def test_fetch_object(test_broker):
    result = await test_broker.publish(
        {"bovine_name": "bovine_name", "object_id": "http://object.example"},
        exchange=processed,
        queue=fetch_object_queue,
        rpc=True,
    )

    assert result == {"some": "object"}



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from unittest.mock import AsyncMock

from . import ProcessorList


async def test_processor_list():
    mock1 = AsyncMock()
    mock2 = AsyncMock()
    mock1.return_value = None

    processor_list = ProcessorList(mock1, mock2)

    item = "item"

    result = await processor_list(item)

    mock1.assert_awaited_once()
    mock2.assert_not_awaited()

    assert result is None
